from machine import Pin, PWM, Timer, UART
import time

uart = UART(1, baudrate=9600, tx=Pin(4), rx=Pin(5))  # Using UART 1 for Pico B
print("Listening for UART messages...")

# Rear Left Pins
RLENA = Pin(13, Pin.OUT)
RLIN1 = Pin(14, Pin.OUT)
RLIN2 = Pin(15, Pin.OUT)
RLencoderPinA = Pin(2, Pin.IN, Pin.PULL_UP)
RLencoderPinB = Pin(3, Pin.IN, Pin.PULL_UP)

# Rear Right Pins
RRENB = Pin(10, Pin.OUT)
RRIN3 = Pin(11, Pin.OUT)
RRIN4 = Pin(12, Pin.OUT)
RRencoderPinA = Pin(0, Pin.IN, Pin.PULL_UP)
RRencoderPinB = Pin(1, Pin.IN, Pin.PULL_UP)

# Front Left Pins
FLENA = Pin(18, Pin.OUT)
FLIN1 = Pin(17, Pin.OUT)
FLIN2 = Pin(16, Pin.OUT)
FLencoderPinA = Pin(26, Pin.IN, Pin.PULL_UP)
FLencoderPinB = Pin(22, Pin.IN, Pin.PULL_UP)

# Front Right Pins
FRENB = Pin(21, Pin.OUT)
FRIN3 = Pin(20, Pin.OUT)
FRIN4 = Pin(19, Pin.OUT)
FRencoderPinA = Pin(27, Pin.IN, Pin.PULL_UP)
FRencoderPinB = Pin(28, Pin.IN, Pin.PULL_UP)

# Encoder variables
RLencoder_position = 0
RRencoder_position = 0
FLencoder_position = 0
FRencoder_position = 0

# PID control variables
Kp = 0.3  # Reduced Kp for less aggressive control
Ki = 0.2  # Reduced Ki
Kd = 0.3  # Reduced Kd
moving_average_size = 5  # Size of the moving average window
rpm_calculation_period = 0.04

# rpm stuff
globalRPM = 30

RLsetpointRPM = 0
RLinputRPM = 0.0
RLoutputPWM = 0.0
RLprevious_error = 0.0
RLintegral = 0.0

RRsetpointRPM = 0
RRinputRPM = 0.0
RRoutputPWM = 0.0
RRprevious_error = 0.0
RRintegral = 0.0

FLsetpointRPM = 0
FLinputRPM = 0.0
FLoutputPWM = 0.0
FLprevious_error = 0.0
FLintegral = 0.0

FRsetpointRPM = 0
FRinputRPM = 0.0
FRoutputPWM = 0.0
FRprevious_error = 0.0
FRintegral = 0.0

# Constants
counts_per_revolution = 1920.0

# Moving average buffer
RLrpm_buffer = []
RRrpm_buffer = []
FLrpm_buffer = []
FRrpm_buffer = []

# PWM setup
RLpwm = PWM(RLENA)
RLpwm.freq(1000)

RRpwm = PWM(RRENB)
RRpwm.freq(1000)

FLpwm = PWM(FLENA)
FLpwm.freq(1000)

FRpwm = PWM(FRENB)
FRpwm.freq(1000)

# Interrupt handler for encoder
def RRencoder(pin):
    global RRencoder_position

    if RRencoderPinA.value() == RRencoderPinB.value():
        RRencoder_position += 1
    else:
        RRencoder_position -= 1

def RLencoder(pin):
    global RLencoder_position

    if RLencoderPinA.value() == RLencoderPinB.value():
        RLencoder_position -= 1
    else:
        RLencoder_position += 1

def FRencoder(pin):
    global FRencoder_position

    if FRencoderPinA.value() == FRencoderPinB.value():
        FRencoder_position += 1
    else:
        FRencoder_position -= 1

def FLencoder(pin):
    global FLencoder_position

    if FLencoderPinA.value() == FLencoderPinB.value():
        FLencoder_position -= 1
    else:
        FLencoder_position += 1

# Attach interrupt to encoder pins
RLencoderPinA.irq(trigger=Pin.IRQ_RISING, handler=RLencoder)
RRencoderPinA.irq(trigger=Pin.IRQ_RISING, handler=RRencoder)
FLencoderPinA.irq(trigger=Pin.IRQ_RISING, handler=FLencoder)
FRencoderPinA.irq(trigger=Pin.IRQ_RISING, handler=FRencoder)

def pid_control():
    global RRprevious_error, RRintegral, RRoutputPWM
    global RLprevious_error, RLintegral, RLoutputPWM
    global FRprevious_error, FRintegral, FRoutputPWM
    global FLprevious_error, FLintegral, FLoutputPWM

    # Rear Right Motor PID
    error = RRsetpointRPM - RRinputRPM
    RRintegral += error
    derivative = error - RRprevious_error
    RRoutputPWM = Kp * error + Ki * RRintegral + Kd * derivative
    RRprevious_error = error

    # Rear Left Motor PID
    error = RLsetpointRPM - RLinputRPM
    RLintegral += error
    derivative = error - RLprevious_error
    RLoutputPWM = Kp * error + Ki * RLintegral + Kd * derivative
    RLprevious_error = error

    # Front Right Motor PID
    error = FRsetpointRPM - FRinputRPM
    FRintegral += error
    derivative = error - FRprevious_error
    FRoutputPWM = Kp * error + Ki * FRintegral + Kd * derivative
    FRprevious_error = error

    # Front Left Motor PID
    error = FLsetpointRPM - FLinputRPM
    FLintegral += error
    derivative = error - FLprevious_error
    FLoutputPWM = Kp * error + Ki * FLintegral + Kd * derivative
    FLprevious_error = error

def set_motor_speed(RLspeed, RRspeed, FLspeed, FRspeed):
    # Rear Right Motor
    if RRspeed > 0:
        RRIN3.value(1)
        RRIN4.value(0)
    elif RRspeed < 0:
        RRIN3.value(0)
        RRIN4.value(1)
        RRspeed = -RRspeed
    else:
        RRIN3.value(0)
        RRIN4.value(0)
    
    # Convert speed percentage to duty cycle (0-65535)
    duty_cycle = int((RRspeed / 100) * 65535)
    RRpwm.duty_u16(duty_cycle)
    
    # Rear Left Motor
    if RLspeed > 0:
        RLIN1.value(0)
        RLIN2.value(1)
    elif RLspeed < 0:
        RLIN1.value(1)
        RLIN2.value(0)
        RLspeed = -RLspeed
    else:
        RLIN1.value(0)
        RLIN2.value(0)
    
    # Convert speed percentage to duty cycle (0-65535)
    duty_cycle = int((RLspeed / 100) * 65535)
    RLpwm.duty_u16(duty_cycle)
    
    # Front Right Motor
    if FRspeed > 0:
        FRIN3.value(1)
        FRIN4.value(0)
    elif FRspeed < 0:
        FRIN3.value(0)
        FRIN4.value(1)
        FRspeed = -FRspeed
    else:
        FRIN3.value(0)
        FRIN4.value(0)
    
    # Convert speed percentage to duty cycle (0-65535)
    duty_cycle = int((FRspeed / 100) * 65535)
    FRpwm.duty_u16(duty_cycle)
    
    # Front Left Motor
    if FLspeed > 0:
        FLIN1.value(0)
        FLIN2.value(1)
    elif FLspeed < 0:
        FLIN1.value(1)
        FLIN2.value(0)
        FLspeed = -FLspeed
    else:
        FLIN1.value(0)
        FLIN2.value(0)
    
    # Convert speed percentage to duty cycle (0-65535)
    duty_cycle = int((FLspeed / 100) * 65535)
    FLpwm.duty_u16(duty_cycle)

while True:
    if uart.any():
        message = uart.read().decode().strip() # remove

        try:
            # Split and convert message to individual RPM values
            device, iRLsetpointRPM, iRRsetpointRPM, iFLsetpointRPM, iFRsetpointRPM, iRLdeg, iRRdeg, iFLdeg, iFRdeg = message.split()

            if device == "m":
                RLsetpointRPM = int(iRLsetpointRPM)
                RRsetpointRPM = int(iRRsetpointRPM)
                FLsetpointRPM = int(iFLsetpointRPM)
                FRsetpointRPM = int(iFRsetpointRPM)
                
                
            else:
                print("not for me")

        except ValueError:
            print("Invalid UART message format. Expected 8 numbers.")
            print(message)

    RRinputRPM = (166.666 * RRencoder_position / counts_per_revolution) / rpm_calculation_period
    RLinputRPM = (166.666 * RLencoder_position / counts_per_revolution) / rpm_calculation_period 
    FRinputRPM = (166.666 * FRencoder_position / counts_per_revolution) / rpm_calculation_period
    FLinputRPM = (166.666 * FLencoder_position / counts_per_revolution) / rpm_calculation_period
    
    print("RR", RRinputRPM, "RL", RLinputRPM, "FR", FRinputRPM, "FL", FLinputRPM)
    
    # Reset encoder positions for the next calculation period
    RRencoder_position = 0
    RLencoder_position = 0
    FRencoder_position = 0
    FLencoder_position = 0
    
    pid_control()
    set_motor_speed(RLoutputPWM, RRoutputPWM, FLoutputPWM, FRoutputPWM)

    # Small delay to prevent overwhelming the CPU
    time.sleep(rpm_calculation_period)
